#include "config.h"
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <string.h>
#include <stdlib.h>
#include "defs.h"
#include "font.h"
#include "tooltip.h"
#include "pixmap.h"
#include "atoms.h"
#include "globals.h"

#define BORDER 4

#define LINE_SPACING 4

static char **CreateStringArray(const char *pText)
{
  /* count the number of \n's */
  char **pRet;
  const char *pIter;
  char *pCopy;
  unsigned int numLines = 1;
  unsigned int i = 0;
  
  for (pIter = pText;*pIter;++pIter)
  {
    if (*pIter == '\n')
    {
      ++numLines;
    }
  }
  /* now allocate structure */
  i = sizeof(char*)*(numLines+1);
  pRet = (char**)malloc(i);
  if (pRet)
  {
    memset(pRet,0,i);
    
    pCopy = strdup(pText);
    if (pCopy)
    {
      i = 0;
      pIter = strtok(pCopy,"\n");
      while (pIter)
      {
	pRet[i++] = strdup(pIter);
	
	pIter = strtok(NULL,"\n");
      }
      free(pCopy);
    }
  }
  return pRet;
}

static void FreeStringArray(char **pArray)
{
  char **pStr;
  for (pStr=pArray;*pStr;pStr++)
  {
    free(*pStr);
  }
  free(pArray);
}

static void RenderTooltip(LPTOOLTIP pTooltip)
{
  char **pStr;
  int y = pTooltip->pFont->ascent + BORDER;

  XSetForeground(pDisp,gGc,PALETTE_BLACK);
  XDrawRectangle(pDisp,pTooltip->wnd,gGc,0,0,
		 pTooltip->width-1,pTooltip->height-1);

  for (pStr=pTooltip->pText;*pStr;++pStr)
  {
    pTooltip->pFont->Render(pTooltip->pFont,pTooltip->wnd,BORDER,y,
			    *pStr,strlen(*pStr));

    y += pTooltip->pFont->ascent + LINE_SPACING;
  }
}

static void CalculateSize(char **pText,unsigned int showFlags,LPFONT pFont,int *x,int *y,unsigned int *width,unsigned int *height)
{
  char **pStr;
  unsigned int tw;
  /* set initial values */
  unsigned int w = 0;
  unsigned int h = 0;
  Window root_ret;
  int dumX,dumY;
  unsigned int root_width,root_height,border_width,depth;
  
  for (pStr=pText;*pStr;pStr++)
  {
    /* calculate this */
    tw = pFont->GetWidth(pFont,*pStr,strlen(*pStr)); 
   
    w = max(w,tw);
    h += pFont->ascent + LINE_SPACING;
  }

  /* ok now set it up so that it fits on the screen */
  
  *width  = w + 2 * BORDER;
  *height = h + 2 * BORDER;

  *x = *x - (*width)/2;
  *y -= (*height + 2);

  if (showFlags == TOOLTIP_SHOW_BOTTOM_LEFT)
  {
    *y += *height + 16;
  }

  XGetGeometry(pDisp,root,&root_ret,&dumX,&dumY,&root_width,&root_height,
	       &border_width,&depth);

  if ((*x + *width) > root_width)
  {
    *x = root_width - *width;
  }
  if ((*y + *height) > root_height)
  {
    *y = root_height - *height;
  }
}

LPTOOLTIP CreateTooltip(const char *pText,int x,int y,unsigned long background,unsigned int showFlags,LPFONT pFont)
{
  XSetWindowAttributes att;
  MWMHints mwm;
  XSizeHints sizeHints;
  LPTOOLTIP pTooltip = (LPTOOLTIP)malloc(sizeof(TOOLTIP));
  if (pTooltip)
  {
    memset(pTooltip,0,sizeof(TOOLTIP));
 
    pTooltip->pText = CreateStringArray(pText);
    if (!pTooltip->pText)
    {
      free(pTooltip);
      return NULL;
    }
    
    pTooltip->pFont = pFont;
    pFont->AddRef(pFont);

    CalculateSize(pTooltip->pText,showFlags,pFont,&x,&y,
		  &pTooltip->width,
		  &pTooltip->height);

    pTooltip->background_pixel = att.background_pixel = background;
    att.event_mask =  ExposureMask;
    att.override_redirect = True;
    pTooltip->wnd = XCreateWindow(pDisp,root,x,y,
				  pTooltip->width,pTooltip->height,0,
				  CopyFromParent,InputOutput,CopyFromParent,
				  CWEventMask | CWBackPixel | CWOverrideRedirect,&att);
    
    /* borderless motif */
    memset(&mwm,0,sizeof(mwm));
    mwm.flags = MWM_HINTS_DECORATIONS;
    XChangeProperty(pDisp,pTooltip->wnd,atom__MOTIF_WM_HINTS,atom__MOTIF_WM_HINTS,32,PropModeReplace,(unsigned char*)&mwm,sizeof(MWMHints)/4);

    sizeHints.flags = PPosition;
    XChangeProperty(pDisp,pTooltip->wnd,XA_WM_NORMAL_HINTS,XA_WM_SIZE_HINTS,32,PropModeReplace,(unsigned char*)&sizeHints,sizeof(XSizeHints)/4);

    XMapWindow(pDisp,pTooltip->wnd);
  }
  return pTooltip;
}

void DestroyTooltip(LPTOOLTIP pTooltip)
{
  XDestroyWindow(pDisp,pTooltip->wnd);
  pTooltip->pFont->Release(pTooltip->pFont);
  FreeStringArray(pTooltip->pText);
  
  free(pTooltip);
}

int ProcessTooltip(LPTOOLTIP pTooltip,XEvent *pEvent)
{
  if (pEvent->xexpose.window == pTooltip->wnd)
  {
    if (pEvent->xexpose.count == 0)
    {
      RenderTooltip(pTooltip);
    }
    return 1;
  }
  return 0;
}
